import json
from zipfile import ZipFile
import os
from typing import Tuple, List
import spacy
from tqdm.auto import tqdm
from hc_nlp.spacy_helpers import correct_entity_boundaries
from hc_nlp import logging

logger = logging.get_logger(__name__)


def load_raw_labelstudio_results(file_name: str) -> dict:
    """
    Load results from a Label Studio zip export.

    Args:
        file_name (str): path to a zip file generated by Label Studio

    Returns:
        dict: see Label Studio docs
    """

    with ZipFile(file_name, "r") as z:
        data = json.loads(z.read("result.json"))

    return data


def load_text_and_annotations_from_labelstudio(
    file_name: str, spacy_model=None, adjust_entity_boundaries=True
) -> List[List[Tuple[str, List[Tuple[int, int, str]]]]]:
    """
    Load text and completed annotations from Label Studio, in the form [("my text", [(1, 3, "ORG"), (4, 10, "PERSON")]), ...].
    `spacy_model` is required for tokenization if `adjust_entity_boundaries` is set to True.

    Args:
        file_name (str): path to a zip file generated by Label Studio
        spacy_model: model used for tokenization
        adjust_entity_boundaries: whether to correct entity boundaries according to token boundaries found by the tokenizer.
            If true, will adjust the start and end of the location of each entity to the closest true start or end.

    Returns:
        List[List[Tuple[str, List[Tuple[int, int, str]]]]]: text, annotations
    """

    docs = load_raw_labelstudio_results(file_name)
    results = []

    for doc in docs:
        text = doc["data"]["text"]

        annotations = []

        for completion in doc["completions"]:
            for annot in completion["result"]:
                annotations.append(
                    (
                        annot["value"]["start"],
                        annot["value"]["end"],
                        annot["value"]["labels"][0],
                    )
                )

        if adjust_entity_boundaries:
            annotations = correct_entity_boundaries(spacy_model, text, annotations)

        results.append((text, annotations))

    return results


def export_text_to_docbin(text: List[str], output_path: str, spacy_model):
    """
    Export list of strings to a spacy DocBin.

    Args:
        text (List[str]): list of text to export
        output_path (str): path to export docbin to. Should end in '.docbin'.
        spacy_model
    """

    if not output_path.endswith(".docbin"):
        logger.warning(
            f"Output path for a DocBin should end in '.docbin'. This will not affect behaviour "
            "but is the recommended extension for a DocBin serialised to disk."
        )

    logger.info("Adding text to DocBin")

    docbin = spacy.tokens.DocBin(store_user_data=True)

    for item in tqdm(text):
        docbin.add(spacy_model(item))

    docbin_data = docbin.to_bytes()

    logger.info("Writing data to file")

    with open(output_path, "wb") as f:
        f.write(docbin_data)



import logging
import sys


def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    stream_handler = logging.StreamHandler(sys.stdout)
    format_string = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    formatter = logging.Formatter(format_string)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    return logger



from hc_nlp import pipeline, constants, io
import spacy
import os

nlp = spacy.load("en_core_web_sm")
nlp_aug = spacy.load("en_core_web_sm")
test_data_path = os.path.join(os.path.dirname(__file__), "2020-11-25-11-43-02.zip")


def test_MapEntityTypes():
    """
    Test that entity type mapping works by creating two pipelines, one of which has the mapper and one of which doesn't.
    """
    mapping = constants.SPACY_TO_HC_ENTITY_MAPPING
    mapping_inputs = set(mapping.keys())

    mapentitytypes = pipeline.MapEntityTypes(nlp, mapping, validate_mapping=True)

    nlp_aug.add_pipe(mapentitytypes)

    data = io.load_text_and_annotations_from_labelstudio(
        test_data_path, spacy_model=nlp, adjust_entity_boundaries=False
    )

    for text, annotations in data[0:100]:
        doc = nlp(text)
        doc_aug = nlp_aug(text)

        doc_entlabels = [ent.label_ for ent in doc.ents]
        doc_aug_entlabels = [ent.label_ for ent in doc_aug.ents]

        assert all(
            [
                ent_aug == mapping[ent] if ent in mapping_inputs else ent_aug == ent
                for ent, ent_aug in zip(doc_entlabels, doc_aug_entlabels)
            ]
        )



"""
Functions to help with Spacy things.
"""

import spacy
import numpy as np
from typing import List, Union


def correct_entity_boundaries(
    spacy_model, text: str, annotations: List[tuple]
) -> List[tuple]:
    """
    Correct start and end positions of entities so that they are on token boundaries, with tokens
    defined by the tokenizer in `spacy_model`.

    Args:
        spacy_model: Spacy model with tokenizer, e.g. the result of spacy.load("en_core_web_sm")
        text (str)
        annotations (List[tuple]): e.g. [(93, 104, 'PERSON'), (113, 129, 'ORG')]. Class names do not 
            need to match the names in the NER component of `spacy_model`.

    Returns:
        List[tuple]: corrected annotations
    """

    # tokens is only needed for debugging
    # tokens = []
    starts = []
    ends = []

    for token in spacy_model(text):
        # tokens.append(token.text)
        starts.append(token.idx)
        ends.append(token.idx + len(token.text))

    new_annotations = []

    for start, end, label in annotations:
        corrected_start = starts[
            min(range(len(starts)), key=lambda i: abs(starts[i] - start))
        ]
        corrected_end = ends[min(range(len(ends)), key=lambda i: abs(ends[i] - end))]

        new_annotations.append((corrected_start, corrected_end, label))

    return new_annotations


def display_manual_annotations(text: str, annotations: List[tuple], **kwargs):
    """
    Use spacy.displacy to display manual annotations created using Label Studio.
    Kwargs are passed to `spacy.displacy.render`.

    Args:
        text (str): [description]
        annotations (List[tuple]): e.g. [(93, 104, 'PERSON'), (113, 129, 'ORG')]
    """

    data = {
        "text": text,
        "ents": [
            {"start": item[0], "end": item[1], "label": item[2]} for item in annotations
        ],
    }

    spacy.displacy.render(data, style="ent", manual=True, **kwargs)


def display_ner_annotations(
    text_or_doc: Union[str, spacy.tokens.Doc], spacy_model=None, **kwargs
):
    """
    Use spacy.display to display annotations created using the NER component of spacy_model or a doc object.
    Kwargs are passed to `spacy.displacy.render`.

    Args:
        text (Union[str, spacy.tokens.Doc]): string or Spacy Doc object. `spacy_model` is not required if Doc object is passed.
        spacy_model: must contain an NER component
        spacy_doc (spacy.tokens.Doc): spacy Doc object
    """

    # assert ("ner" in spacy_model.pipe_names) or ("entity_ruler" in spacy_model.pipe_names)

    if isinstance(text_or_doc, spacy.tokens.Doc):
        spacy.displacy.render(text_or_doc, style="ent", **kwargs)
    elif spacy_model:
        spacy.displacy.render(spacy_model(text_or_doc), style="ent", **kwargs)
    else:
        raise ValueError(
            "Please provide either a Spacy Doc or both a string and a spacy model."
        )



### PATTERNS FOR FINDING ENTITIES OF SPECIFIC TYPES

DATE_PATTERNS = [
    {"label": "DATE", "pattern": [{"SHAPE": "dddd"}, {"ORTH": "-"}, {"SHAPE": "dddd"}]},  # 1984 - 1990 | 1984-1990
    {"label": "DATE", "pattern": [{"ORTH": "c."}, {"SHAPE": "dddd"}]},  # c. 1200
    {"label": "DATE", "pattern": [{"TEXT": {"REGEX": r"c\.\d{3,4}"}}]},  # c.1200
    {"label": "DATE", "pattern": [{"TEXT": {"REGEX": r"\d{1,2}/\d{1,2}/(\d{4}|\d{2})"}}]},  # 03/12/2000
    {"label": "DATE", "pattern": [{"TEXT": {"REGEX": r"\d{1,2}\.\d{1,2}\.(\d{4}|\d{2})"}}]},  # 03.12.2000
    {"label": "DATE", "pattern": [{"SHAPE": "dd"}, {"ORTH": "-"}, {"SHAPE": "dd"}, {"ORTH": "-"}, {"SHAPE": "dddd"}]},  # 03-12-2000
    {"label": "DATE", "pattern": [{"SHAPE": "d"}, {"ORTH": "-"}, {"SHAPE": "dd"}, {"ORTH": "-"}, {"SHAPE": "dddd"}]},  # 3-12-2000
    {"label": "DATE", "pattern": [{"SHAPE": "dd"}, {"ORTH": "-"}, {"SHAPE": "d"}, {"ORTH": "-"}, {"SHAPE": "dddd"}]},  # 03-1-2000
    {"label": "DATE", "pattern": [{"SHAPE": "d"}, {"ORTH": "-"}, {"SHAPE": "d"}, {"ORTH": "-"}, {"SHAPE": "dddd"}]},  # 3-1-2000
    {"label": "DATE", "pattern": [{"SHAPE": "dddd"}, {"ORTH": "to"}, {"SHAPE": "dddd"}]},  # 1805 to 1860
]

### USEFUL CONSTANTS

ORDINALS = [
    "1st", "2nd", "3rd", "4th", "5th", "6th", "7th", "8th", "9th", "10th",
    "11th", "12th", "13th", "14th", "15th", "16th", "17th", "18th", "19th", "20th", "21st",
    "first", "second", "third", "fourth", "fifth", "sixth", "seventh", "eighth", "ninth",
    "tenth", "eleventh", "twelfth", "thirteenth", "fourteenth", "fifteenth", "sixteenth",
    "seventeenth", "eighteenth", "nineteenth", "twentieth", "twenty-first"
]

### SPACY TO HERITAGECONNECTOR ENTITY MAPPING
### This allows us to map multiple Spacy NER classes to one class, in the case that we don't need the 
### detail of specific classes. Also includes identity mappings. `set(SPACY_TO_HC_ENTITY_MAPPING.values())`
### is the complete set of entities are using for the Heritage Connector.
SPACY_TO_HC_ENTITY_MAPPING = {
    "PERSON": "PERSON",
    "ORG": "ORG",
    "NORP": "NORP",
    "GPE": "LOC",
    "LOC": "LOC",
    "PRODUCT": "OBJECT",
    "WORK_OF_ART": "OBJECT",
    "EVENT": "EVENT",
    "DATE": "DATE",
}



from hc_nlp import spacy_helpers
import spacy

nlp = spacy.load("en_core_web_sm")


def test_correct_entity_boundaries():
    text = "There are 59 bulbs in Hungary."
    doc = nlp(text)
    spacy_annotations = [(ent.start_char, ent.end_char, ent.label_) for ent in doc.ents]

    # these annotations are off compared to the correct Spacy ones
    incorrect_annotations = [(11, 12, "CARDINAL"), (22, 27, "GPE")]

    corrected_annotations = spacy_helpers.correct_entity_boundaries(
        nlp, text, incorrect_annotations
    )

    assert corrected_annotations == spacy_annotations



import spacy
from spacy.pipeline import EntityRuler
import time
from typing import List
from hc_nlp import constants, logging

logger = logging.get_logger(__name__)


class ThesaurusMatcher:
    """
    The ThesaurusMatcher lets you add spans to `Doc.ents` using exact phrase
    matches from an imported phrasebook (Thesaurus). It can be combined with 
    the statistical `EntityRecognizer` to boost accuracy, or used on its own 
    to implement a purely rule-based entity recognition system. After 
    initialization, the component is typically added to the pipeline using 
    `nlp.add_pipe`.
    """

    def __init__(
        self,
        nlp,
        thesaurus_path: str,
        case_sensitive: bool,
        overwrite_ents: bool = False,
    ):
        """
        Initialise the ThesaurusMatcher. `thesaurus_path` must point to a .jsonl
        file with each line in the following format (with the `id` key optional):

        ```
        {
            "label": "<entity label>",
            "pattern": "<text to match>",
            "id": "<ID which the above text unambiguously refers to. Optional.>",
        }
        ```

        Args:
            nlp: spacy model
            thesaurus_path (str): path to the thesaurus
            case_sensitive (bool): [description]
        """
        if case_sensitive:
            self.ruler = EntityRuler(nlp)
        else:
            self.ruler = EntityRuler(nlp, phrase_matcher_attr="LOWER")

        self.nlp = nlp
        self.thesaurus_path = thesaurus_path
        self._add_thesaurus_to_ruler()
        self.ruler.overwrite = overwrite_ents

    def _add_thesaurus_to_ruler(self):
        """
        Load thesaurus from disk and add to self.ruler
        """
        logger.info(f"Loading thesaurus from {self.thesaurus_path}")
        other_pipes = [p for p in self.nlp.pipe_names if p != "tagger"]

        start = time.time()
        with self.nlp.disable_pipes(*other_pipes):
            self.ruler.from_disk(self.thesaurus_path)

        end = time.time()
        logger.info(f"{len(self.ruler)} term thesaurus imported in {int(end-start)}s")

    def __call__(self, doc: spacy.tokens.Doc) -> spacy.tokens.Doc:
        """
        Inherits from EntityRuler behaviour.
        """
        return self.ruler(doc)


class EntityFilter:
    """
    The EntityFilter filters out any entities in `Doc.ents` that aren't likely to be
    entities based on their form. It uses the following rules to determine this, which
    can be switched off or changed when initialising the EntityFilter:
    - under `max_token_length` -> not entity
    - contains 3+ consecutive digits -> maybe entity (could be a date)
    - all lowercase or all UPPERCASE -> not entity

    These rules are applied so that if any of the tokens in an entity span don't look 
    like entities, the whole span will be removed from `Doc.ents`.
    """

    def __init__(
        self,
        max_token_length: int = 1,
        remove_all_lower: bool = True,
        remove_all_upper: bool = True,
    ):
        """
        Initialise the EntityFilter.

        Args:
            max_token_length (int, optional): Entities with tokens with length less
                than or equal to this will be removed from Doc.ents. Defaults to 1.
            remove_all_lower (bool, optional): Entities with one or more lowercase 
                token are removed. Defaults to True.
            remove_all_upper (bool, optional): Entities with one or more uppercase
                token are removed. Defaults to True.
        """
        self.max_token_length = max_token_length
        self.remove_all_lower = remove_all_lower
        self.remove_all_upper = remove_all_upper

    def _is_unlikely_entity(self, token: spacy.tokens.Token) -> bool:
        """
        Returns True if a token is likely not an entity, and False otherwise.
        """
        if "ddd" in token.shape_:
            # tokens with 3+ consecutive digits could be years, so possibly DATE entities
            return False

        elif len(token) <= self.max_token_length:
            return True

        # UPPERCASE and lowercase tokens assumed not to be entities
        elif token.is_lower and self.remove_all_lower:
            return True

        elif token.is_upper and self.remove_all_upper:
            return True

        return False

    def __call__(self, doc: spacy.tokens.Doc) -> spacy.tokens.Doc:
        """
        Filter out entities which contain one or more token that doesn't look like
        an entity.
        """
        likely_entity = []
        for ent in doc.ents:
            likely_entity.append(
                any([not self._is_unlikely_entity(tok) for tok in ent])
            )

        doc.ents = [ent for idx, ent in enumerate(doc.ents) if likely_entity[idx]]

        return doc


class PatternMatcher:
    """
    An EntityRuler object initiated with a pattern. Used for built-in `hc_nlp`
    matchers.
    """

    def __init__(self, nlp, patterns: List[dict]):
        """
        Initialise the PatternMatcher.

        Args:
            nlp : Spacy model
            patterns (List[dict]): for the EntityRuler. See https://spacy.io/usage/rule-based-matching#entityruler 
        """
        self.ruler = EntityRuler(nlp)
        self.ruler.add_patterns(patterns)

    def __call__(self, doc: spacy.tokens.Doc) -> spacy.tokens.Doc:
        """
        Inherits from EntityRuler behaviour.
        """
        return self.ruler(doc)


class DateMatcher(PatternMatcher):
    def __init__(self, nlp):
        super().__init__(nlp, constants.DATE_PATTERNS)

    def _add_centuries_to_doc(self, doc: spacy.tokens.Doc) -> spacy.tokens.Doc:
        """
        Adds dates with the format "... nth century" to `Doc.ents`. Does this by finding the word
        'century' or 'centuries', checking that the previous word is an ordinal, and then returning all 
        the immediate children of the token 'century'. It then checks for occurrences of "nth (and/or/to) 
        mth centuries", as well as "AD" or "BC" after the word century/centuries.

        Args:
            doc (spacy.tokens.Doc)

        Returns:
            spacy.tokens.Doc
        """
        for idx, token in enumerate(doc):
            if token.lower_ in ["century", "centuries"]:
                if (doc[idx - 1].lower_ in constants.ORDINALS) or all(
                    [
                        string in constants.ORDINALS
                        for string in doc[idx - 1].lower_.split("-")
                    ]
                ):
                    try:
                        first_child = next(token.children)
                    except Exception:  #  noqa: E722
                        # if the token has no children, use the ordinal token as first_child
                        first_child = doc[idx - 1]

                    # allow "nth (and|to|or) mth" century
                    if (doc[first_child.i - 1].lower_ in ["and", "to", "or"]) and (
                        doc[first_child.i - 2].lower_ in constants.ORDINALS
                    ):
                        # go back to the first child of "nth"
                        start = next(doc[first_child.i - 2].children).i

                        # if the child is after the 'nth' token, use the token instead of its child
                        if start > doc[first_child.i - 2].i:
                            start = doc[first_child.i - 2].i
                    else:
                        start = first_child.i

                    # print([c for c in doc[first_child.i - 2].children])

                    end = idx + 1

                    # add on 'AD' or 'BC' if present
                    if idx != len(doc) - 1:
                        if doc[idx + 1].text.upper() in ["AD", "BC"]:
                            end += 1

                    date_entity = spacy.tokens.Span(doc, start, end, label="DATE")
                    doc.ents = list(doc.ents) + [date_entity]
        return doc

    def __call__(self, doc: spacy.tokens.Doc) -> spacy.tokens.Doc:
        """
        Detects centuries then patterns from `constants.DATE_PATTERNS`.

        Args:
            doc (spacy.tokens.Doc)

        Returns:
            spacy.tokens.Doc
        """
        doc = self._add_centuries_to_doc(doc)
        doc = self.ruler(doc)

        return doc


class MapEntityTypes:
    def __init__(
        self,
        nlp,
        mapping: dict = constants.SPACY_TO_HC_ENTITY_MAPPING,
        validate_mapping: bool = True,
    ):
        entities_missing_from_mapping = set(nlp.entity.labels) - set(mapping.keys())

        if validate_mapping and (len(entities_missing_from_mapping) > 0):
            logger.warning(
                f"The following entity labels from the Spacy model are not in the provided mapping: {', '.join(list(entities_missing_from_mapping))}. They will not be changed."
            )

        self.mapping = mapping
        self.nlp = nlp

    def __call__(self, doc: spacy.tokens.Doc) -> spacy.tokens.Doc:
        """
        Replace entities in `Doc.ents` with new entities based on the mapping specified when initialising
        the class instance.

        Args:
            doc (spacy.tokens.Doc

        Returns:
            spacy.tokens.Doc
        """
        new_ents = []
        for ent in list(doc.ents):
            new_ent = spacy.tokens.Span(
                doc, ent.start, ent.end, label=self.mapping.get(ent.label_, ent.label_)
            )
            new_ents.append(new_ent)

        doc.ents = new_ents

        return doc



import sys

sys.path.append("..")

from hc_nlp import io
import spacy
import os

nlp = spacy.load("en_core_web_sm")
test_data_path = os.path.join(os.path.dirname(__file__), "2020-11-25-11-43-02.zip")


def test_load_raw_labelstudio_results():
    data = io.load_raw_labelstudio_results(test_data_path)

    assert isinstance(data, list)
    assert isinstance(data[0], dict)

    assert all([item in data[0].keys() for item in ["completions", "data", "id"]])


def test_load_text_and_annotations_from_labelstudio():
    """
    Data should be in the format [[text, [(start1, end1, label1), (start2, end2, label2), ...]]]
    """
    data = io.load_text_and_annotations_from_labelstudio(
        test_data_path, spacy_model=nlp, adjust_entity_boundaries=False
    )

    # test that all 'text' parts can be converted to string
    assert all([str(item[0]) for item in data])

    # test that all starts and ends are integers
    assert [all([isinstance(x, int) for x in i[0:1]]) for item in data for i in item[1]]

    # test that all labels are strings
    assert all([isinstance(i[2], str) for item in data for i in item[1]])



import spacy
from spacy.gold import GoldParse
from spacy.scorer import Scorer
from typing import List, Tuple
from collections import Counter

from hc_nlp.io import load_text_and_annotations_from_labelstudio
from hc_nlp import logging

logger = logging.get_logger(__name__)


def test_ner(
    spacy_model,
    results_set: str = None,
    examples: List[List[Tuple[str, List[Tuple[int, int, str]]]]] = None,
) -> dict:
    """
    Return precision, recall and F-score for a Spacy NER model based on a set of gold-standard
    labels returned by Label Studio. One of `results_set` and `examples` should be provided.

    Args:
        spacy_model: model with NER component
        results_set (str, optional): name of results set from labelling/export folder
        examples (List[List[Tuple[str, List[Tuple[int, int, str]]]]], optional): from `io.load_text_and_annotations_from_labelstudio`

    Returns:
        dict: keys ents_p; ents_r; ents_f; ents_per_type
    """

    assert "ner" in spacy_model.pipe_names

    if (results_set and examples) or (not results_set and not examples):
        raise ValueError("Please provide exactly one of `results_set` and `examples`.")

    elif results_set:
        examples = load_text_and_annotations_from_labelstudio(results_set, spacy_model)

    scorer = Scorer()
    for input_, annot in examples:
        doc_gold_text = spacy_model.make_doc(input_)
        gold = GoldParse(doc_gold_text, entities=annot)
        pred_value = spacy_model(input_)
        scorer.score(pred_value, gold)

    entity_measures = ["ents_p", "ents_r", "ents_f", "ents_per_type"]
    ent_results = {k: scorer.scores[k] for k in entity_measures}

    label_count = _count_labels(examples)
    total_support = 0
    labels_with_support = []

    for label, count in label_count.items():
        if label in ent_results["ents_per_type"].keys():
            ent_results["ents_per_type"][label]["support"] = count
            labels_with_support.append(label)
            total_support += count

        else:
            logger.warn(
                f"Label {label} not in the model. To get a list of labels in a Spacy model `nlp` run `nlp.entity.labels`."
            )

    ent_results["support"] = total_support
    ent_results["labels_missing_from_annotations"] = list(
        set(ent_results["ents_per_type"].keys()) - set(labels_with_support)
    )
    ent_results["ents_per_type"] = {
        k: ent_results["ents_per_type"][k] for k in labels_with_support
    }

    return ent_results


def _count_labels(examples: List[List[Tuple[str, List[Tuple[int, int, str]]]]]) -> dict:
    """
    Return the number of examples present for each label in examples.

    Args:
        examples (List[List[Tuple[str, List[Tuple[int, int, str]]]]]): ([text, [(annotation_start, annotation_end, annotation_label), ...]])

    Returns:
        dict: {annotation_label: label_count, ...}
    """

    label_list = [item[2] for ex in examples for item in ex[1]]

    return dict(Counter(label_list))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hc-nlp",
    version="0.2.0",
    author="Science Museum Group",
    description="Heritage Connector NLP",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TheScienceMuseum/heritage-connector-nlp",
    download_url="https://github.com/TheScienceMuseum/heritage-connector-nlp/archive/v0.1.0.tar.gz",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pandas>=1.1.4",
        "spacy>=2.3.0,<2.4.0",
        "label-studio==0.8.1",
        "jupyterlab==2.2.9",
        "tqdm>=4.10.0,<5.0.0",
        "seaborn==0.11.0",
    ],
    packages=["hc_nlp"],
)
